(ns demo.org
  (:require [clojure.string :as str]))


(def ESCAPE "\n")

(def BLANK  0)
(def META  1)
(def META_OTHER  2)
(def HEADER  5)
(def BOLD  10)
(def ITALIC  11)
(def UNDERLINED  12)
(def VERBATIM  13)
(def LIST  20)
(def TEXT  21)
(def IMAGE  22)
(def LINK  23)
(def CAPTION  24)
(def BULLET  25)
(def SOURCE  50)
(def EXAMPLE  51)
(def RESULTS  52)
(def COMMENT  53)
(def TABLE  54)

(def METADATA  ["TITLE"
                "AUTHOR"
                "EMAIL"
                "DESCRIPTION"
                "KEYWORDS"
                "FILETAGS"
                "DATE"
                "HTML_DOCTYPE"
                "SETUPFILE"])

(def t_META   #"^[#]\+(?:TITLE|AUTHOR|EMAIL|DESCRIPTION|KEYWORDS|FILETAGS|DATE|HTML_DOCTYPE|SETUPFILE)\:")
(def t_BLANK_LINE  #"^\s*$")
(def t_COMMENT_BEGIN #"^\#\+BEGIN_COMMENT")
(def t_COMMENT_END #"^\#\+END_COMMENT")
(def t_EXAMPLE_BEGIN #"^\#\+BEGIN_EXAMPLE")
(def t_EXAMPLE_END  #"^\#\+END_EXAMPLE")
(def t_SRC_BEGIN  #"^\#\+BEGIN_SRC\s+")
(def t_SRC_END  #"^\#\+END_SRC")
(def t_TABLE_START  #"^\s*\|")
(def t_TABLE_END  #"^(?!\s*\|).*$")
(def t_RESULTS_START  #"^\#\+RESULTS\:")
(def t_CAPTIONS  #"^\#\+CAPTION:")
(def t_NAME  #"^\#\+NAME:")
; t_IMG  #"^\[\[(\w|\.|-|_|/)+\]\]$"
(def t_IMG  #"^\[\[")
(def t_IMG_END  #"\]\]")
(def t_RESULTS_END  #"^\s*$")
(def t_END_LABELS  #"^(?!\[|\#).*")
(def t_BULLET_START  #"^\s*[\+|\-|0-9\.]")
(def t_BULLET_END  #"^(?!\s*[\+|\-|0-9\.]).*$")

(def t_HEADER  #"^\*+")
(def t_META_OTHER  #"^[#]\+[A-Z\_]+\:")

(def TYPE_SINGLE  0)
(def TYPE_BLOCK 1)
(def TYPE_ATTRIBUTE  2)

(defn TokenStruct [values]
  (merge {:start "" :end false :type TYPE_SINGLE :start_pos 2 :end_pos nil :count 0 :key ""} values))

(def token-map {:token nil :value nil :attrs nil})

(defn token
  ([token-type value] (token token-type value nil))
  ([token-type value attrs]
   {:token token-type :value (or value "") :attrs attrs}))

(def TOKENS [[:META (TokenStruct {:start t_META, :end_pos -1 :key-fn #(clojure.string/join (drop 2 (butlast %)))})],
             [:COMMENT (TokenStruct {:start t_COMMENT_BEGIN, :end t_COMMENT_END, :type TYPE_BLOCK, :end_pos -1})],
             [:EXAMPLE (TokenStruct {:start t_EXAMPLE_BEGIN, :end t_EXAMPLE_END, :type TYPE_BLOCK, :end_pos -})],
             [:IMAGE (TokenStruct {:start t_IMG, :end_pos -2})],
             [:CAPTION (TokenStruct {:start t_CAPTIONS, :type TYPE_ATTRIBUTE, :key "caption"})],
             [:br (TokenStruct {:start t_BLANK_LINE, :end_pos -1})],
             [:SOURCE (TokenStruct {:start t_SRC_BEGIN, :end t_SRC_END})],
             [:TABLE (TokenStruct {:start t_TABLE_START, :end t_TABLE_END, :start_pos 0})],
             [:BULLET (TokenStruct {:start t_BULLET_START, :end t_BULLET_END, :start_pos 0})],
             [:RESULTS (TokenStruct {:start t_RESULTS_START, :end t_RESULTS_END})],
             [:HEADER (TokenStruct {:start t_HEADER, :start_pos 1, :count 0})],
             [:META_OTHER (TokenStruct {:start t_META_OTHER, :start_pos 2, :end_pos -1})]])


(def rm-ns (comp keyword name))

(defn mk-key [k t pos]
  (case (name k)
    "HEADER" (keyword (str (name k) pos))
    (rm-ns k)))

(clojure.string/join (drop 2 (butlast "#+title:")))
(re-find t_META "#+TITLE: abc12345def")
(re-find t_HEADER "** test")
(str/split  "#+TITLE: abc12345def" t_META)
(def t
  "#+TITLE: abc12345def
#+DESCRIPTION: test descriptioon

* title
paragraph
text
** header test
hi there
")

(defn re-split [rx line]
  (let [matches (re-find rx line)
        pos (if (string? matches)
              (count matches)
              (count (or (str (first matches)) matches)))
        remainder (subs line pos)]
    [matches pos remainder]))

(defn match-token [line]
  (->> TOKENS
       (reduce (fn [m [k v]]
                 (let [s (re-find (:start v) line)
                       pos (if (string? s) (count s) (count (or (str (first s)) s)))
                       v (subs line pos)]
                   (when (not (nil? s))
                     (reduced (remove nil? [(mk-key k v pos) (if (= "" v) nil v)]))))) [])
       (#(or % [:p line]))))


(defn parse-line [text]
  (loop [line (str/split-lines text)
         r []]
    (if (empty? line)
      r
      (recur (next line)
             (if (= :HEADER1 (first (last r)))
               (conj (last r) (match-token (first line)))
               (conj r (match-token (first line))))))))

(defn parse-flat [text]
  (->> (str/split-lines text)
       (map (fn [a] (match-token a)))))

(parse-line t )
(parse-flat t )

(defn to->html [dsl]
  (map #(apply conj [(case (first %)
                       :TITLE :h1
                       :HEADER1 :h1
                       :HEADER2 :h2
                       :br :br
                       :SOURCE :pre
                       (first %))] (rest %)) dsl))

(defn parse->to-hiccup [org-document]
  (to->html (parse-flat org-document)))

(parse->to-hiccup t)
(parse->to-hiccup "")

(map #(clojure.set/rename-keys (parse-line t) {:HEADER :h1}) (parse-line t ))
